#ifndef TEMPSET_H
#define TEMPSET_H

#include <map>
#include <algorithm>
using namespace std;

template<class T> class TempSet;

/// Iteratore per TempSet
/**
 * L'iteratore  descritto da un puntatore al TempSet a cui si riferisce e da un iteratore
 * ad una multimap<T,int> che  contenuta nel TempSet.<br>
 * Sono realizzati i metodi minimi che ci servono: l'incremento, il confronto, la deferenziazione.
 */
template<class T> class TSiter{
	private:
		typedef multimap<T,int>::iterator ait;	///< tipo dell'iteratore ausiliario
		typedef const T& reference_type;	///< tipo reference
		TempSet<T>* cont;	///< container corrente
		ait curr;	///< itaratore ausiliario corrente

	public:
		/// costruttore banale
		TSiter() {};
		/// costruttore utile
		/**
		 * \param x il TempSet di riferimento
		 * \param v un iteratore alla multimap del TempSet di riferimento
		 */
		TSiter(TempSet<T>& x, ait v): cont(&x), curr(v) {};
		/// incremento (serve per scandire un intervallo)
		TSiter operator++();
		/// uguaglianza
		template<class C> friend bool operator==(const TSiter<C>& sx, const TSiter<C>& dx);
		/// deferenziazione
		reference_type operator*() const;
};

/// Multiset temporeneo
/**
 * La classe TempSet fornisce una versione temporanea del multiset.<br>
 * Insieme all'oggetto da inserire viene passata un'eventuale data di scadenza.<br>
 * Il TempSet si occuper di ignorare gli oggetti scaduti e di rimuoverli
 * di tanto in tanto dalla memoria.<br>
 * La classe  formata da una multimap<T,int> usata per memorizzare le coppie (oggetto,scadenza).
 */
template<class T> class TempSet{
	private:
		/// l'incremento dell'iteratore deve essere friend
		friend TSiter<T> TSiter<T>::operator++();
		typedef TSiter<T> iterator;	///<tipo iteratore
		typedef multimap<T,int>::iterator ait;	///< tipo iteratore ausiliario
		/// multimap per memorizzare i dati
		multimap<T,int> TS;
		/// contatore di find e equal_range
		int count;
		/// numero di find da far passare prima di una pulizia
		static const int clean=50;
		/// toglie gli oggetti scaduti dalla memoria
		void purge();
	public:
		/// iteratore a dopo la fine del TempSet
		iterator end();
		/// cerca un oggetto
		/**
		 * \return l'iteratore all'oggetto trovato o end() se non trovato
		 */
		iterator find(const T&);
		/// cerca un intervallo di oggetti
		/**
		 * \return una coppia di iteratori al primo e subito dopo
		 * l'ultimo degli oggetti trovati. entrambi end() in caso di ricerca fallita
		 */
		pair<iterator, iterator> equal_range(const T&);
		/// inserisce un oggetto con eventuale scadenza
		void insert(const T&, int=0);
};

/// predicato controllo scadenza
/**
 * usato da TempSet<T>::purge() per controllare quali dati
 * rimuovere perch scaduti
 */
template<class C> class old : public unary_function< pair<const C,int>, bool>{
	public:
		/// controlla scadenza
		bool operator()(pair<const C,int> cop) {
			return cop.second<100;
		}
};

template<class T> TSiter<T>::reference_type TSiter<T>::operator*() const{
	return curr->first;
}

template<class T> TSiter<T> TSiter<T>::operator++(){
	ait it;
	for(it=++curr; it!=cont->TS.end() && old<T>()(*it); ++it);
	curr=it;
	
}

template<class T> bool operator==(const TSiter<T>& sx, const TSiter<T>& dx){
	return sx.cont==dx.cont && sx.curr==dx.curr;
}

template<class T> TempSet<T>::iterator TempSet<T>::end(){
	return TSiter<T>(*this, TS.end());
}

template<class T> TempSet<T>::iterator TempSet<T>::find(const T& what){
	if (++count%clean==0)
		purge();
	pair<ait,ait> coppia=TS.equal_range(what);
	ait ret;
	for(ret=coppia.first; old<T>()(*ret) && ret!=coppia.second; ++ret);
	return ret==coppia.second? TSiter<T>(*this, TS.end()): TSiter<T>(*this, ret);
}

template<class T> void TempSet<T>::purge(){
	multimap< T,int > purged;
	remove_copy_if(TS.begin(), TS.end(), inserter(purged,purged.end()), old<T>() );
	TS=purged;		 
}

template<class T> pair<TempSet<T>::iterator,TempSet<T>::iterator> TempSet<T>::equal_range(const T& what){
	if (++count%clean==0)
		purge();
	pair<ait,ait> coppia=TS.equal_range(what);

	ait f;
	for(f=coppia.first; old<T>()(*f) && f!=coppia.second; ++f);
	iterator fi,la;
	if (f==coppia.second){
		fi=la=TSiter<T>(*this, TS.end());
	} else {
		fi=TSiter<T>(*this, f);
		ait l=coppia.second;
		while (l!=TS.end() && old<T>()(*l))
			++l;
		la=TSiter<T>(*this, l);
	}

	return make_pair(fi,la);
}

template<class T> void TempSet<T>::insert(const T& what, int time=0){
	TS.insert(make_pair(what,time));
}

#endif
